#include <iostream>			//cout cin
#include <bitset>
#include <stdio.h>
#include <errno.h>			//Error msgs
#include <string.h>			//Strings
#include <fstream>      	//File handling
#include <linux/kernel.h>	//Hw clock
#include <sys/socket.h>		//UDP handling
#include <sys/types.h>
#include <sys/sysinfo.h>	//Uptime
#include <sys/time.h>		//Set clock
#include <arpa/inet.h>
#include <netinet/in.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>			//Use sleep
#include <mavlink.h>		//Mavlink protocol
#include "sa_api.h"			//Custom header SA44B
#include <cstdio>       	//C type iostream
#include <cstdlib>      	//C type stdlib.h
#include <cmath>        	//Math handling
#include <unistd.h>     	//Fuctions of c/c++
#include <ctime>        	//Timestamp
#include <limits.h>
#include <termios.h>		//Serial port
#include <fcntl.h>			//Serial port
#include <signal.h>			//Handling ctrl+c routine

using namespace std;

#define BUFFER_LENGTH 2104 //buffer size max mavlink message

//**********************FILES*******************************************************/
ofstream data_log,event_log,error_log; //Logfile(s)

//**********************FUNCTIONS**************************************************/
uint64_t microsSinceEpoch();
void intHandler(int);
int open_serial();
bool open_sa44b();
void perform_sweep();
long get_uptime();

//**********************GLOBALS**************************************************/
bool active = true;

//Serial to usb parameters:
const char *portname = "/dev/ttyUSB-USB2SER";	//Port for mavlink com
int BAUDRATE = B57600;							//Baudrate of mavlink com

//SA44B parameters:
int DEVICE;										//SA44B
int SWEEPS = 8;								//Sweep numbers
int RESBW = 4000;								//Resolution bandwidth in Hz (not rec <3kHz)
double FREQUENCY = 221208;						//Measuring frequency in kHz
double SPAN = 1000;								//Measuring span in Hz

//**********************MAIN***********************************************/
int main()
{
	//Program variables:
	signal(SIGINT, intHandler);					//ctrl+c routine
	bool relax = true;
	enum {communicate, measure, ending} mode;	//00:com, 01:meas, 10:end, 11:not used
	mode = communicate;							//Start with communication
	bool synced = false;						//Mode: not synced
	int timeout = 0;							//
	uint8_t buf[BUFFER_LENGTH];					//Buffer for serial receive
	ssize_t recsize;							//Variable for serial receive

	//Mavlink variables:
	mavlink_message_t msg;						//Contains the mavlink message
	mavlink_status_t status;					//Holds the status
	mavlink_system_time_t sys_time;				//Holds the time variables

	//Time variables:
	time_t new_sys_time;
	time_t old_sys_time;
	time_t time_offset;
	time_t msg_boot_time;
	time_t boot_time;

	//Serial port:
	system("/home/odroid/mavlink/examples/linux/usb_off"); 	//Switch ftdi module to serial reading
	usleep(10);
	int ser_dev = open_serial();							//Open serial for reading
	usleep(100);											//Wait

	//Create log file(s):
	data_log.open("data.txt");								//Log data
	event_log.open("events.txt");							//Log program events
	error_log.open("error.txt");							//Log errors

	//**********************MAIN****************************************/
	while(active)
	{
		//Communication mode:
		if(mode == communicate)																		//Tries to read packages from the pixhawk
		{
			memset(buf, 0, BUFFER_LENGTH);															//clear buffer
			recsize = read(ser_dev, buf, BUFFER_LENGTH);											//Read serial port
			if(recsize > 0)																			//Something received
			{
				for(int i=0; i < recsize; i++)														//Read whole buffer
				{
					if(mavlink_parse_char(MAVLINK_COMM_0, buf[i], &msg, &status))					//If message is mavlink
					{
						switch(msg.msgid)															//Switch on msgid number
						{
							case MAVLINK_MSG_ID_SYSTEM_TIME: 										//#2
								mavlink_msg_system_time_decode(&msg, &sys_time);					//Decode time info
				/*DEBUG*/		event_log/*cout*/ << time(0) << ": "<< "time_unix_usec:" << sys_time.time_unix_usec << "\nsys_time.time_boot_ms:" << sys_time.time_boot_ms << '\n'; /*DEBUG*/
				/*DEBUG*/		//if(sys_time.time_boot_ms > 20000){mode = measure;}/*DEBUG*/
								if(not synced)
								{
									if(sys_time.time_unix_usec == 0 || sys_time.time_boot_ms == 0)	//Something wrong with the time
									{
										break;														//Invalid data read again
									}

									else															//Time is valid, store:
									{
										//Store time variables:
										new_sys_time = sys_time.time_unix_usec/1e6;					//pixhawk sys time s
										old_sys_time = time(0);										//linux sys time s
										msg_boot_time = sys_time.time_boot_ms;						//pixhawk boot time ms
									}

									//Try to set new system clock time:
									if(stime(&new_sys_time) == -1)
									{
										//cout << "error set time failed\n";
										error_log << "error set time failed\n" << "Not synced\n";
										synced = false;
										exit(EXIT_FAILURE);
									}

									//Try to set new hw clock time:
									//Write to log:
									event_log << "Unix time msg: "  	<< new_sys_time
										  << "\nBoot time msg: "	<< msg_boot_time
										  << "\nSystem time is: "<< old_sys_time
										  << "\nDiff time is: "  << time_offset
										  << '\n';

									//Set hardware clock:
									system("hwclock -w");
									event_log << "New system clock set!\n" << "New HW clock set!\n" << "Sync complete\n";

									//Sync complete:
									synced = true;
									mode = measure; //
								}//End not synced*/
								//mode = measure;//DEBUGGING
								break;
							case MAVLINK_MSG_ID_HEARTBEAT: 											//#0
								event_log/*cout*/ << time(0) << ": "<< "MSG:" << int(msg.msgid) << " Heartbeat\n";break;
/*							case MAVLINK_MSG_ID_COMMAND_ACK:break;									//#77
							case MAVLINK_MSG_ID_MISSION_REQUEST:break;								//#40
							case MAVLINK_MSG_ID_MISSION_COUNT:break;								//#44
							case MAVLINK_MSG_ID_RAW_IMU:break;										//#27
							case MAVLINK_MSG_ID_SCALED_IMU2:break;									//#116
							case MAVLINK_MSG_ID_SCALED_PRESSURE:break;								//#29
							case MAVLINK_MSG_ID_SYS_STATUS:break;									//#01
							case MAVLINK_MSG_ID_POWER_STATUS:break;									//#125
							case MAVLINK_MSG_ID_MISSION_CURRENT:break;								//#42
							case MAVLINK_MSG_ID_GPS_RAW_INT:break;									//#24
							case MAVLINK_MSG_ID_NAV_CONTROLLER_OUTPUT:break;						//#62
							case MAVLINK_MSG_ID_GLOBAL_POSITION_INT:break;							//#33
							case MAVLINK_MSG_ID_LOCAL_POSITION_NED:break;							//#32
							case MAVLINK_MSG_ID_SERVO_OUTPUT_RAW:break;								//#36
							case MAVLINK_MSG_ID_RC_CHANNELS_RAW:break;								//#35
							case MAVLINK_MSG_ID_RC_CHANNELS:break;									//#65
							case MAVLINK_MSG_ID_ATTITUDE:break;										//#30
							case MAVLINK_MSG_ID_VFR_HUD:break;										//#74
							case MAVLINK_MSG_ID_TERRAIN_REPORT:break;								//#136
							case MAVLINK_MSG_ID_VIBRATION:break;									//#241
							case MAVLINK_MSG_ID_STATUSTEXT:break;									//#253
							case MAVLINK_MSG_ID_MISSION_ITEM_REACHED:break;							//#46
							case MAVLINK_MSG_ID_PARAM_VALUE:break;									//#22
							case MAVLINK_MSG_ID_HOME_POSITION:break;								//#242
							case MAVLINK_MSG_ID_MISSION_ITEM:break;									//#39
							case MAVLINK_MSG_ID_SET_MODE:break;		 								//#11						*/
							default:
								//printf("%02X ", msg.msgid);										//Prints out non known msg
								break;
						}																			//End switch
					}																				//End if mavlink msg
				}																					//End for loop received
			}																						//End if received something
			else																					//Not received anything
			{
				if(timeout < 60)//Not connected in 1 min, abort
				{
					timeout++;
					sleep(1);//Delay a little
				}

				else
				{
					error_log/*cout*/ << time(0) << ": "<< "Link not active!\n";
					mode = ending;
				}
			}
		}
		
		//Measurement mode:
		else if(mode == measure)
		{
			if(relax)
			{	//Sets the port first				
				close(ser_dev); //Close the pixhawk link
				usleep(10);
				system("/home/odroid/mavlink/examples/linux/serial_off");//Set up for usb comm				
				usleep(10);//Wait
				relax = not open_sa44b();
			}

			perform_sweep();//Do a sweep
			mode = ending;//Stops measuring and ends the program
		}

		//Not connected yet
		else if(mode == ending)																	//Ending routine
		{
			active = false;																		//Stop loop and close
			boot_time = get_uptime()*1e3;														//linux boot time ms
			event_log/*cout*/ << time(0) << ": "<< "Ending at boot time: "  << boot_time << "ms\nSystem time: " << time(0) << "s\n";
		}
		
		usleep(10); 																			//Sleep(justeres til den treffer nok målinger i SA44B)
	}//End while active loop

   	//Close application:
	event_log/*cout*/ << "Mission complete!\n";  	
	memset(buf, 0, BUFFER_LENGTH);	//Clear buffer
	saCloseDevice(DEVICE);																		//Close device
	system("/home/odroid/mavlink/examples/linux/usb_off"); 									//Switch ftdi module to serial reading
	data_log.close();																			//Close log files:
	event_log.close();
	error_log.close();
	system("/home/odroid/mavlink/examples/linux/move_log.sh");
	return 0;
}																								//End main()

//**********************FUNCTIONS******************************************************/
//Abort routine handler ctrl + c
void intHandler(int abort_signal)
{
	active = false;
}

//Open serial port for Mavlink and Pixhawk
int open_serial()
{
	struct termios tty;
	int sc = open(portname, O_RDONLY | O_NOCTTY | O_NDELAY);									//Opens the usb port for serial reading
	event_log/*cout*/ << "Opening port:" << portname << '\n';
	if(sc < 0)
	{
		error_log/*cout*/ << time(0) << ": "<< "error " << errno << " opening " << portname << ": " << strerror(errno) << '\n';
		return -1;
	}
	else
	{
		fcntl(sc, F_SETFL, FNDELAY);
	}
	
	event_log/*cout*/ << time(0) << ": "<< "Port:" << portname << " opened successfull\n";//Log confirm
	
	//Set parameters for serial interface:	
	memset (&tty, 0, sizeof tty);
    if (tcgetattr (sc, &tty) != 0)
	{
    	error_log/*cout*/ << time(0) << ": error " << errno << " from tcgetattr\n";
		return -1;
    }
    cfsetospeed (&tty, BAUDRATE);
    cfsetispeed (&tty, BAUDRATE);
    tty.c_cflag = (tty.c_cflag & ~CSIZE) | CS8;													//Use 8-bit chars
    tty.c_iflag &= ~IGNBRK;																		//Ignore break
    tty.c_lflag = 0;																			//
    tty.c_oflag = 0;																			//
    tty.c_cc[VMIN]  = 0;																		//
    tty.c_cc[VTIME] = 5;																		//
    tty.c_iflag &= ~(IXON | IXOFF | IXANY);														//shut off xon/xoff ctrl
    tty.c_cflag |= (CLOCAL | CREAD);															//ignore modem controls, enable reading
    tty.c_cflag &= ~(PARENB | PARODD);															//shut off parity
    tty.c_cflag |= 0; 																			//no parity
    tty.c_cflag &= ~CSTOPB;
    tty.c_cflag &= ~CRTSCTS;
    if (tcsetattr (sc, TCSANOW, &tty) != 0)														//check for error
    {
    	error_log/*cout*/ << time(0) << ": error " << errno << " from tcsetattr\n";												//notify error
        return -1;																				//return error
    }
    return sc;																					//return device number
}

//Open the SA44B for communication
bool open_sa44b()
{ 
	if(saOpenDevice(&DEVICE) == saNoError)		//If no errors configure device for sweeps:
	{
		event_log/*cout*/ << time(0) << ": "<< "Device opened\n";
		event_log/*cout*/ << time(0) << ": "<< "Config Acquisition: "
			<< saGetErrorString(saConfigAcquisition(DEVICE, SA_MIN_MAX, SA_LOG_SCALE))		//Type of sampling and data type
			<< "\nConfig Center Span: "
			<< saGetErrorString(saConfigCenterSpan(DEVICE, FREQUENCY*1e3, SPAN))		//Center frequency and span
			<< "\nConfig Level: "
			<< saGetErrorString(saConfigLevel(DEVICE, -30.0))	 							//Assumed input signal level dBm
			<< "\nConfig Sweep Coupling: "
			<< saGetErrorString(saConfigSweepCoupling(DEVICE, RESBW, RESBW, false))	//Resolution bw, video bw and a boolean
			<< "\nInitiate: "
			<< saGetErrorString(saInitiate(DEVICE, SA_SWEEPING, 0))							//Enter Sweeping mode
			<< '\n';
	}
	
	else
	{
		error_log/*cout*/ << time(0) << ": "<< "Device error!\n" 
			<< saGetErrorString(saOpenDevice(&DEVICE)) 
			<< "\nTerminating at t=" 
			<< time(0) 
			<< '\n';
		exit(EXIT_FAILURE);
		return false;//May select exit or false return
	}
	
	event_log/*cout*/ << time(0) << ": "<< "Executing "	<< SWEEPS << " sweeps"
			 << " at frequency " 	<< FREQUENCY<< "kHz" 
			 << " with a span of " 	<< SPAN 	<< "Hz"
			 << " and a RBW of " 	<< RESBW 	<< "Hz\n";
	return true;										//Device ready
}

//Perform sweep
void perform_sweep()
{
  	// Get some info about the sweep
   	int sweepLen;
	double startFreq, binSize;

   	if(saQuerySweepInfo(DEVICE, &sweepLen, &startFreq, &binSize) != saNoError)
	{
		error_log/*cout*/ << time(0) << ": "<< "Error Query Sweep Info\n" << time(0) << '\n';
	}

   	// Allocate arrays for the sweep data:
   	float *min = new float[sweepLen];
   	float *max = new float[sweepLen];

	for(int k=1; k <= SWEEPS; k++)
	{
		data_log/*cout*/ << time(0) << "," << saGetErrorString(saGetSweep_32f(DEVICE, min, max)) << ", ";      	//Sweep is captured

		//Peak search
   	 	int maxIdx = 0;
   	 	for(int i=1; i < sweepLen; i++)
		{
   		 	(max[i] > max[maxIdx]) ? maxIdx = i : 0;
   		}

   		//Display peak frequency, amplitude for each sweep.
   		double peakFreq = 1.0e-6 * (startFreq + binSize * maxIdx);
		data_log/*cout*/ << peakFreq << " ,MHz, " << max[maxIdx] << ",dBm" << "\n";
	}
	usleep(100);

   	//Stop sweeping and return the device to "idle"
   	saAbort(DEVICE);
	event_log/*cout*/ << time(0) << ": "<< "End of sweep\n";

  	//Avoid memory leaks
	delete [] min;
	delete [] max;
}

//Get the boot time of Odroid
long get_uptime()
{
	struct sysinfo s_info;
	int error = sysinfo(&s_info);
	if(error != 0)
	{
		error_log/*cout*/ << time(0) << ": "<< "Error uptime, code = " << error;
	}
	return s_info.uptime;
}

//**********************END************************/
